public inbox for linux-kernel@vger.kernel.org
 help / color / mirror / Atom feed
From: Caleb Sander Mateos <csander@purestorage.com>
To: Ming Lei <ming.lei@redhat.com>, Jens Axboe <axboe@kernel.dk>,
	Shuah Khan <shuah@kernel.org>
Cc: linux-block@vger.kernel.org, linux-kselftest@vger.kernel.org,
	linux-kernel@vger.kernel.org,
	Stanley Zhang <stazhang@purestorage.com>,
	Uday Shankar <ushankar@purestorage.com>,
	Caleb Sander Mateos <csander@purestorage.com>
Subject: [PATCH 13/20] ublk: optimize ublk_user_copy() on daemon task
Date: Tue, 16 Dec 2025 22:34:47 -0700	[thread overview]
Message-ID: <20251217053455.281509-14-csander@purestorage.com> (raw)
In-Reply-To: <20251217053455.281509-1-csander@purestorage.com>

ublk user copy syscalls may be issued from any task, so they take a
reference count on the struct ublk_io to check whether it is owned by
the ublk server and prevent a concurrent UBLK_IO_COMMIT_AND_FETCH_REQ
from completing the request. However, if the user copy syscall is issued
on the io's daemon task, a concurrent UBLK_IO_COMMIT_AND_FETCH_REQ isn't
possible, so the atomic reference count dance is unnecessary. Check for
UBLK_IO_FLAG_OWNED_BY_SRV to ensure the request is dispatched to the
sever and obtain the request from ublk_io's req field instead of looking
it up on the tagset. Skip the reference count increment and decrement.
Commit 8a8fe42d765b ("ublk: optimize UBLK_IO_REGISTER_IO_BUF on daemon
task") made an analogous optimization for ublk zero copy buffer
registration.

Signed-off-by: Caleb Sander Mateos <csander@purestorage.com>
---
 drivers/block/ublk_drv.c | 23 ++++++++++++++++++-----
 1 file changed, 18 insertions(+), 5 deletions(-)

diff --git a/drivers/block/ublk_drv.c b/drivers/block/ublk_drv.c
index 042df4de9253..a0fbabd49feb 100644
--- a/drivers/block/ublk_drv.c
+++ b/drivers/block/ublk_drv.c
@@ -180,11 +180,11 @@ struct ublk_io {
 	/*
 	 * The number of uses of this I/O by the ublk server
 	 * if user copy or zero copy are enabled:
 	 * - UBLK_REFCOUNT_INIT from dispatch to the server
 	 *   until UBLK_IO_COMMIT_AND_FETCH_REQ
 	 * - 1 for each inflight ublk_ch_{read,write}_iter() call
 	 * - 1 for each io_uring registered buffer not registered on task
 	 * The I/O can only be completed once all references are dropped.
 	 * User copy and buffer registration operations are only permitted
 	 * if the reference count is nonzero.
 	 */
@@ -2644,10 +2644,11 @@ ublk_user_copy(struct kiocb *iocb, struct iov_iter *iter, int dir)
 	struct ublk_queue *ubq;
 	struct request *req;
 	struct ublk_io *io;
 	unsigned data_len;
 	bool is_integrity;
+	bool on_daemon;
 	size_t buf_off;
 	u16 tag, q_id;
 	ssize_t ret;
 
 	if (!user_backed_iter(iter))
@@ -2670,13 +2671,24 @@ ublk_user_copy(struct kiocb *iocb, struct iov_iter *iter, int dir)
 
 	if (tag >= ub->dev_info.queue_depth)
 		return -EINVAL;
 
 	io = &ubq->ios[tag];
-	req = __ublk_check_and_get_req(ub, q_id, tag, io);
-	if (!req)
-		return -EINVAL;
+	on_daemon = current == READ_ONCE(io->task);
+	if (on_daemon) {
+		/* On daemon, io can't be completed concurrently, so skip ref */
+		if (!(io->flags & UBLK_IO_FLAG_OWNED_BY_SRV))
+			return -EINVAL;
+
+		req = io->req;
+		if (!ublk_rq_has_data(req))
+			return -EINVAL;
+	} else {
+		req = __ublk_check_and_get_req(ub, q_id, tag, io);
+		if (!req)
+			return -EINVAL;
+	}
 
 	if (is_integrity) {
 		struct blk_integrity *bi = &req->q->limits.integrity;
 
 		data_len = bio_integrity_bytes(bi, blk_rq_sectors(req));
@@ -2697,11 +2709,12 @@ ublk_user_copy(struct kiocb *iocb, struct iov_iter *iter, int dir)
 		ret = ublk_copy_user_integrity(req, buf_off, iter, dir);
 	else
 		ret = ublk_copy_user_pages(req, buf_off, iter, dir);
 
 out:
-	ublk_put_req_ref(io, req);
+	if (!on_daemon)
+		ublk_put_req_ref(io, req);
 	return ret;
 }
 
@@ -180,11 +180,11 @@ struct ublk_io {
 	 * - UBLK_REFCOUNT_INIT from dispatch to the server
 	 *   until UBLK_IO_COMMIT_AND_FETCH_REQ
-	 * - 1 for each inflight ublk_ch_{read,write}_iter() call
+	 * - 1 for each inflight ublk_ch_{read,write}_iter() call not on task
 	 * - 1 for each io_uring registered buffer not registered on task
 	 * The I/O can only be completed once all references are dropped.
 static ssize_t ublk_ch_read_iter(struct kiocb *iocb, struct iov_iter *to)
 {
-- 
2.45.2


  parent reply	other threads:[~2025-12-17  5:35 UTC|newest]

Thread overview: 40+ messages / expand[flat|nested]  mbox.gz  Atom feed  top
2025-12-17  5:34 [PATCH 00/20] ublk: add support for integrity data Caleb Sander Mateos
2025-12-17  5:34 ` [PATCH 01/20] block: validate pi_offset integrity limit Caleb Sander Mateos
2025-12-18  8:53   ` Christoph Hellwig
2025-12-17  5:34 ` [PATCH 02/20] block: validate interval_exp " Caleb Sander Mateos
2025-12-18  8:53   ` Christoph Hellwig
2025-12-17  5:34 ` [PATCH 03/20] blk-integrity: take const pointer in blk_integrity_rq() Caleb Sander Mateos
2025-12-19 14:16   ` Ming Lei
2025-12-17  5:34 ` [PATCH 04/20] ublk: add integrity UAPI Caleb Sander Mateos
2025-12-22 14:26   ` Ming Lei
2025-12-22 15:09     ` Caleb Sander Mateos
2025-12-23  1:51       ` Ming Lei
2025-12-17  5:34 ` [PATCH 05/20] ublk: move ublk flag check functions earlier Caleb Sander Mateos
2025-12-22 14:30   ` Ming Lei
2025-12-17  5:34 ` [PATCH 06/20] ublk: support UBLK_PARAM_TYPE_INTEGRITY in device creation Caleb Sander Mateos
2025-12-22 14:47   ` Ming Lei
2025-12-22 15:35     ` Caleb Sander Mateos
2025-12-23  1:58       ` Ming Lei
2025-12-17  5:34 ` [PATCH 07/20] ublk: set UBLK_IO_F_INTEGRITY in ublksrv_io_desc Caleb Sander Mateos
2025-12-22 14:48   ` Ming Lei
2025-12-17  5:34 ` [PATCH 08/20] ublk: add ublk_copy_user_bvec() helper Caleb Sander Mateos
2025-12-22 14:52   ` Ming Lei
2025-12-22 15:37     ` Caleb Sander Mateos
2025-12-17  5:34 ` [PATCH 09/20] ublk: split out ublk_user_copy() helper Caleb Sander Mateos
2025-12-22 14:58   ` Ming Lei
2025-12-17  5:34 ` [PATCH 10/20] ublk: inline ublk_check_and_get_req() into ublk_user_copy() Caleb Sander Mateos
2025-12-26  2:10   ` Ming Lei
2025-12-17  5:34 ` [PATCH 11/20] ublk: move offset check out of __ublk_check_and_get_req() Caleb Sander Mateos
2025-12-26  2:15   ` Ming Lei
2025-12-17  5:34 ` [PATCH 12/20] ublk: implement integrity user copy Caleb Sander Mateos
2025-12-26  2:38   ` Ming Lei
2025-12-17  5:34 ` Caleb Sander Mateos [this message]
2025-12-26  2:51   ` [PATCH 13/20] ublk: optimize ublk_user_copy() on daemon task Ming Lei
2025-12-17  5:34 ` [PATCH 14/20] selftests: ublk: add utility to get block device metadata size Caleb Sander Mateos
2025-12-17  5:34 ` [PATCH 15/20] selftests: ublk: add kublk support for integrity params Caleb Sander Mateos
2025-12-17  5:34 ` [PATCH 16/20] selftests: ublk: implement integrity user copy in kublk Caleb Sander Mateos
2025-12-17  5:34 ` [PATCH 17/20] selftests: ublk: support non-O_DIRECT backing files Caleb Sander Mateos
2025-12-17  5:34 ` [PATCH 18/20] selftests: ublk: add integrity data support to loop target Caleb Sander Mateos
2025-12-17  5:34 ` [PATCH 19/20] selftests: ublk: add integrity params test Caleb Sander Mateos
2025-12-17  5:34 ` [PATCH 20/20] selftests: ublk: add end-to-end integrity test Caleb Sander Mateos
2025-12-18 16:54 ` (subset) [PATCH 00/20] ublk: add support for integrity data Jens Axboe

Reply instructions:

You may reply publicly to this message via plain-text email
using any one of the following methods:

* Save the following mbox file, import it into your mail client,
  and reply-to-all from there: mbox

  Avoid top-posting and favor interleaved quoting:
  https://en.wikipedia.org/wiki/Posting_style#Interleaved_style

* Reply using the --to, --cc, and --in-reply-to
  switches of git-send-email(1):

  git send-email \
    --in-reply-to=20251217053455.281509-14-csander@purestorage.com \
    --to=csander@purestorage.com \
    --cc=axboe@kernel.dk \
    --cc=linux-block@vger.kernel.org \
    --cc=linux-kernel@vger.kernel.org \
    --cc=linux-kselftest@vger.kernel.org \
    --cc=ming.lei@redhat.com \
    --cc=shuah@kernel.org \
    --cc=stazhang@purestorage.com \
    --cc=ushankar@purestorage.com \
    /path/to/YOUR_REPLY

  https://kernel.org/pub/software/scm/git/docs/git-send-email.html

* If your mail client supports setting the In-Reply-To header
  via mailto: links, try the mailto: link
Be sure your reply has a Subject: header at the top and a blank line before the message body.
This is a public inbox, see mirroring instructions
for how to clone and mirror all data and code used for this inbox